@@ -1,24 +1,40 @@
 from __future__ import annotations
 
+import ast
+import functools
+import itertools
 import json
 from collections import Counter
 from copy import copy, deepcopy
-from itertools import chain, combinations
+from itertools import chain
 from typing import Callable, Dict, List, Optional, Tuple, Union
 
 import numpy as np
 from numpy.random import rand
+from py_expression_eval import Parser
 from pyDOE import lhs
-from scipy.special import logit
 from sobol_seq import i4_sobol_generate
 
-from .._exception import ConstraintEvaluationError
+from ..utils.exception import ConstraintEvaluationError
 from .node import Node
 from .samplers import SCMC
 from .variable import Bool, Discrete, Integer, Ordinal, Real, Subset, Variable
 
 __authors__ = "Hao Wang"
 
+_reduce = lambda iterable: functools.reduce(lambda a, b: a + b, iterable)
+_get_var = lambda c: Parser().parse(c).variables()[0]
+
+# TODO: add sampling method for the tree structure
+
+
+def _get_val(condition):
+    c = ast.parse(condition).body[0].value.comparators[0]
+    if hasattr(c, "n"):
+        return c.n
+    if hasattr(c, "s"):
+        return c.s
+
 
 class SearchSpace:
     """Search Space Base Class
@@ -55,9 +71,14 @@ class SearchSpace:
             the corresponding discrete variable serves as the dictionary key.
     """
 
-    _supported_types = (Real, Integer, Ordinal, Discrete, Bool)
+    _supported_types = (Real, Integer, Ordinal, Discrete, Bool, Subset)
 
-    def __init__(self, data: List[Variable], random_seed: int = None, hierarchy: dict = None):
+    def __init__(
+        self,
+        data: List[Variable],
+        random_seed: int = None,
+        structure: Union[dict, List[Node]] = None,
+    ):
         """Search Space
 
         Parameters
@@ -75,14 +96,12 @@ class SearchSpace:
 
         self.random_seed: int = random_seed
         self._set_data(data)
+        self._set_structure(structure)
         SearchSpace.__set_type(self)
-        self.structure = hierarchy
-        self.__set_conditions()
 
     @property
     def var_name(self):
         return self._var_name
-        # return (self._var_name[0] if self.dim == 1 else self._var_name)
 
     @var_name.setter
     def var_name(self, var_name):
@@ -152,38 +171,36 @@ class SearchSpace:
                 for i, k in enumerate(idx):
                     data[k].name = _names[i]
 
-    @property
-    def structure(self):
-        return self._structure
-
-    @structure.setter
-    def structure(self, hierarchy: dict):
-        if hierarchy:
-            self._structure = Node.from_dict(hierarchy)
-        else:
-            pass
+    def _set_structure(self, structure: Union[dict, List[Node]] = None):
+        if structure is None:
+            _structure = dict()
+        # a list of tree/nodes
+        elif isinstance(structure, list) and all([isinstance(t, Node) for t in structure]):
+            _structure = dict()
+            for t in structure:
+                _structure.update(t.to_dict())
+        elif isinstance(structure, dict):
+            _structure = structure
+        # scan for all conditions defined in variables
+        for var in self.data:
+            if not hasattr(var, "conditions") or var.conditions is None:
+                continue
+            # TODO: support more dependent variables in the condition
+            key = var.conditions["vars"][0]
+            if key not in var.conditions["vars"]:
+                raise ValueError(f"variable {var} not in {self}")
+            _structure.setdefault(key, []).append(
+                {"name": var.name, "condition": var.conditions["string"]}
+            )
+        self.structure = [t.remove(self.var_name, invert=True) for t in Node.from_dict(_structure)]
+        self.structure = [t for t in self.structure if t]
 
     @staticmethod
     def __set_type(obj: SearchSpace) -> SearchSpace:
         _type = np.unique(obj.var_type)
-        if len(_type) == 1:
-            obj.__class__ = eval(_type[0] + "Space")
-        else:
-            obj.__class__ = SearchSpace
+        obj.__class__ = eval(_type[0] + "Space") if len(_type) == 1 else SearchSpace
         return obj
 
-    def __set_conditions(self):
-        # TODO: perhaps implement a `conditions` property (list of all conditions)
-        # TODO: to validate the conditions specified in variables
-        # TODO: add conditions when the prefix of some variables appears (conditional parameters)
-        for var in self.data:
-            if var.conditions is not None:
-                pre, _, __ = var.name.rpartition(".")
-                if pre:
-                    for i, var_name in enumerate(var.var_in_conditions):
-                        if not var_name.startswith(pre):
-                            var.var_in_conditions[i] = pre + "." + var_name
-
     def _set_data(self, data):
         """Sanity check on the input data and set the auxiliary variables"""
         self._check_data(data)
@@ -206,9 +223,12 @@ class SearchSpace:
                 self.__dict__[attr_mask] = mask
                 self.__dict__[attr_id] = np.nonzero(mask)[0]
 
-            self.categorical_id = np.r_[self.discrete_id, self.ordinal_id, self.bool_id]
+            self.categorical_id = np.r_[
+                self.discrete_id, self.ordinal_id, self.bool_id, self.subset_id
+            ]
             self.categorical_mask = np.bitwise_or(
-                self.bool_mask, np.bitwise_or(self.discrete_mask, self.ordinal_mask)
+                self.bool_mask,
+                np.bitwise_or(self.discrete_mask, self.ordinal_mask, self.subset_mask),
             )
 
     def _set_levels(self):
@@ -241,6 +261,9 @@ class SearchSpace:
             out = data
         elif isinstance(data, list):
             out = SearchSpace(data, self.random_seed)
+            # backwards compatibility
+            if hasattr(self, "structure"):
+                getattr(out, "_set_structure")(self.structure)
         return out
 
     def __setitem__(self, index, value):
@@ -297,7 +320,13 @@ class SearchSpace:
         # NOTE: the random seed of `self` has the priority
         random_seed = self.random_seed if self.random_seed else space.random_seed
         data = deepcopy(self.data) + space.data
-        return SearchSpace(data, random_seed)
+        if hasattr(self, "structure"):
+            structure = [t.deepcopy() for t in self.structure] + [
+                t.deepcopy() for t in space.structure
+            ]
+        else:
+            structure = {}
+        return SearchSpace(data, random_seed, structure)
 
     def __radd__(self, space) -> SearchSpace:
         return self.__add__(space)
@@ -315,7 +344,11 @@ class SearchSpace:
         _res = set(self.var_name) - set(space.var_name)
         _index = [self.var_name.index(_) for _ in _res]
         data = [copy(self.data[i]) for i in range(self.dim) if i in _index]
-        return SearchSpace(data, random_seed)
+        cs = SearchSpace(data, random_seed)
+        # backwards compatibility
+        if hasattr(self, "structure"):
+            getattr(cs, "_set_structure")(self.structure)
+        return cs
 
     def __rsub__(self, space) -> SearchSpace:
         return self.__sub__(space)
@@ -342,6 +375,8 @@ class SearchSpace:
             a copy of replicated `self`
         """
         data = [deepcopy(var) for _ in range(max(1, int(N))) for var in self.data]
+        # TODO: this is not working yet..
+        # structure = [t.deepcopy() for _ in range(max(1, int(N))) for t in self.structure]
         obj = SearchSpace(data, self.random_seed)
         obj.__class__ = type(self)
         return obj
@@ -376,6 +411,13 @@ class SearchSpace:
             msg += str(var) + "\n"
         return msg
 
+    def pprint(self):
+        if self.structure:
+            for root in self.structure:
+                root.pprint(data={k: self[k] for k in self.var_name})
+        else:
+            print(self.__str__())
+
     def filter(self, keys: List[str], invert=False) -> SearchSpace:
         """filter a search space based on a list of variable names
 
@@ -400,14 +442,16 @@ class SearchSpace:
             return args[0]
 
         assert isinstance(args[0], SearchSpace)
-        data = list(chain.from_iterable([deepcopy(_.data) for _ in args]))
-        return SearchSpace(data)
+        data = list(chain.from_iterable([deepcopy(cs.data) for cs in args]))
+        structure = [t.deepcopy() for cs in args for t in cs.structure]
+        return SearchSpace(data, structure=structure)
 
     def pop(self, index: int = -1) -> Variable:
         value = self.data.pop(index)
         self._set_data(self.data)
+        if hasattr(self, "structure"):
+            self._set_structure(self.structure)
         SearchSpace.__set_type(self)
-        self.__set_conditions()
         return value
 
     def remove(self, index: Union[int, str]) -> SearchSpace:
@@ -423,13 +467,15 @@ class SearchSpace:
 
         self.data.pop(_index)
         self._set_data(self.data)
-        self.__set_conditions()
+        # backwards compatibility
+        if hasattr(self, "structure"):
+            self._set_structure(self.structure)
         return SearchSpace.__set_type(self)
 
     def update(self, space: SearchSpace) -> SearchSpace:
         """Update the search space based on the var_name of the input search space,
         which behaves similarly to the dictionary update. Please note its difference
-        to ``self.__add__``
+        to ``self.__add__``. This function will not update the structure of the search space.
 
         Parameters
         ----------
@@ -600,33 +646,36 @@ class SearchSpace:
 
             N = range(int(v["N"])) if "N" in v else range(1)
             default_value = v["defualt"] if "default" in v else None
-            if v["type"] in ["r", "real"]:  # real-valued parameter
-                precision = v["precision"] if "precision" in v else None
-                scale = v["scale"] if "scale" in v else "linear"
-                _vars = [
-                    Real(
-                        bounds,
-                        name=k,
-                        default_value=default_value,
-                        precision=precision,
-                        scale=scale,
-                    )
-                    for _ in N
-                ]
-            elif v["type"] in ["i", "int", "integer"]:  # integer-valued parameter
-                _vars = [
-                    Integer(bounds, name=k, default_value=default_value, step=v.pop("step", 1))
-                    for _ in N
-                ]
-            elif v["type"] in ["o", "ordinal"]:  # ordinal parameter
-                _vars = [Ordinal(bounds, name=k, default_value=default_value) for _ in N]
-            elif v["type"] in ["c", "cat"]:  # category-valued parameter
-                _vars = [Discrete(bounds, name=k, default_value=default_value) for _ in N]
-            elif v["type"] in ["s", "subset"]:  # subset parameter
-                _vars = [Subset(bounds, name=k, default_value=default_value) for _ in N]
-            elif v["type"] in ["b", "bool"]:  # Boolean-valued
-                _vars = [Bool(name=k, default_value=default_value) for _ in N]
-            variables += _vars
+            try:
+                if v["type"] in ["r", "real"]:  # real-valued parameter
+                    precision = v["precision"] if "precision" in v else None
+                    scale = v["scale"] if "scale" in v else "linear"
+                    _vars = [
+                        Real(
+                            bounds,
+                            name=k,
+                            default_value=default_value,
+                            precision=precision,
+                            scale=scale,
+                        )
+                        for _ in N
+                    ]
+                elif v["type"] in ["i", "int", "integer"]:  # integer-valued parameter
+                    _vars = [
+                        Integer(bounds, name=k, default_value=default_value, step=v.pop("step", 1))
+                        for _ in N
+                    ]
+                elif v["type"] in ["o", "ordinal"]:  # ordinal parameter
+                    _vars = [Ordinal(bounds, name=k, default_value=default_value) for _ in N]
+                elif v["type"] in ["c", "cat"]:  # category-valued parameter
+                    _vars = [Discrete(bounds, name=k, default_value=default_value) for _ in N]
+                elif v["type"] in ["s", "subset"]:  # subset parameter
+                    _vars = [Subset(bounds, name=k, default_value=default_value) for _ in N]
+                elif v["type"] in ["b", "bool"]:  # Boolean-valued
+                    _vars = [Bool(name=k, default_value=default_value) for _ in N]
+                variables += _vars
+            except:
+                print(param)
         return SearchSpace(variables)
 
     @classmethod
@@ -646,11 +695,30 @@ class SearchSpace:
         with open(file, "r") as f:
             return cls.from_dict(json.load(f))
 
-    def pprint(self):
-        print(str(self))
+    def get_unconditional_subspace(self) -> List[Tuple[dict, SearchSpace]]:
+        """get all unconditional subspaces"""
         if self.structure:
-            for root in self.structure:
-                root.pprint()
+            # all variables in the conditional structure
+            _var = _reduce([t.get_all_name() for t in self.structure])
+            # remaining variables not affected by conditions
+            isolated_var = [self[v] for v in set(self.var_name) - set(_var)]
+            out, d = list(), dict()
+            # all paths in the conditional tree
+            paths = [list(root.get_all_path().items()) for root in self.structure]
+            idx = [list(range(len(_))) for _ in paths]
+            # get all combinations of paths from all trees
+            for item in itertools.product(*idx):
+                condition = _reduce([paths[i][k][0] for i, k in enumerate(item)])
+                variables = _reduce([paths[i][k][1] for i, k in enumerate(item)])
+                d[condition] = variables
+            # create all unconditional subspaces
+            for condition, var in d.items():
+                key = {_get_var(c): _get_val(c) for c in condition}
+                out.append((key, SearchSpace(isolated_var + [self[v] for v in var])))
+            # TODO: consider the case where selector/conditioning variable has other values
+        else:
+            out = [({}, self)]
+        return out
 
 
 class RealSpace(SearchSpace):
@@ -692,6 +760,7 @@ class RealSpace(SearchSpace):
             X[:, i] = var.round(X[:, i])
         return X
 
+    # TODO: implement a transformation class/function and make this generic
     def to_linear_scale(self, X: Union[np.ndarray, List[List]]) -> np.ndarray:
         X = np.atleast_2d(X).astype(float)
         assert X.shape[1] == self.dim
